
/******************************************************************************
* MODULE     : input.cc
* DESCRIPTION: mathemagix shell input and completion functions
* COPYRIGHT  : (C) 2005  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <shell.hh>
#ifdef SUPPORTS_RL
#include <readline/readline.h>
#include <readline/history.h>
#endif // SUPPORTS_RL

namespace mmx {
using std::cout;
using std::cin;

#ifndef SUPPORTS_RL
static char rl_in[1000];
static char*
readline (char* prompt) {
  cout << prompt;
  cin.getline (rl_in, 1000);
  if (!cin.good()) return NULL;
  return (char*) ((void*) rl_in);
}
static void
add_history (char* s) { (void) s; }
#endif // SUPPORTS_RL
  
#ifdef SUPPORTS_RL

static char*
duplicate (char *s) {
  // strdup does not work on some platforms; I don't know why
  char* r= (char*) malloc (strlen (s) + 1);
  strcpy (r, s);
  return r;
}

/* Generator function for command completion. STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
static char *
completion_generator (const char* text, int state) {
  static int text_len;
  static list current_list;
  string name;
  char *tmp, *copiedname;
  int cmp;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state) {
    current_list = entries (symbol_table);
    text_len = strlen (text);
  }
  
  /* Return the next name which partially matches from the command list. */
  while (!nil(current_list)) {
    name = as_string (car (current_list[0]));
    current_list = cdr (current_list);
    tmp = as_charp (name);
    cmp = strncmp (tmp, text, text_len);
    mmx_free_charp (tmp);
    if (cmp == 0) {
      tmp = as_charp (name);
      copiedname = duplicate (tmp); // Alloc memory for copiedname with malloc
      mmx_free_charp (tmp);
      return copiedname; // Readline frees the strings when it has finished with them.
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* Attempt to complete on the contents of TEXT.  START and END bound
   the region of rl_line_buffer that contains the word to complete.
   TEXT is the word to complete. We can use the entire contents of
   rl_line_buffer in case we want to do some simple parsing.  Return
   the array of matches, or NULL if there aren't any. */
static char **
complete (const char *text, int start, int end) {
  char **matches;

# if defined(RL_VERSION_MAJOR) && RL_VERSION_MAJOR >= 4
  rl_completion_append_character = '\0';
# endif
  matches = (char **)NULL;
  matches = rl_completion_matches(text, completion_generator);
  return (matches);
}
#endif // SUPPORTS_RL

string
shell_terminal_input (int& line_counter) {
  char* line;
#ifdef SUPPORTS_RL
  if (completion_mode) {
    rl_attempted_completion_function = complete;
  }
#endif
  line_counter++;
  if (quiet_mode)
    line = readline (NULL);
  else {
    string prompt= as_string (line_counter) * "> ";
    char* _prompt= as_charp (prompt);
    line = readline (_prompt);
    mmx_free_charp (_prompt);
  }
  if (line==NULL) {
    cout << "\n";
    exit(0);
  }
  if (line[0] != '\0') add_history (line);
  return line;
}

} // namespace mmx
