/******************************************************************************
* MODULE     : texmacs.cc
* DESCRIPTION: mathemagix texmacs interface and completion functions
* COPYRIGHT  : (C) 2005  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <symbol.hh>
#include <shell.hh>

namespace mmx {
using std::cout;
using std::cin;

#define DATA_BEGIN   ((char) 2)
#define DATA_END     ((char) 5)
#define DATA_ESCAPE  ((char) 27)
#define DATA_COMMAND '\20'

// Completion
  
static void
output_completion(string text, list matches) {
  cout << DATA_BEGIN << "scheme:(tuple ";
  cout << '\"' << text << '\"';
  while (!nil(matches)) {
    if (N(as_string(car(matches)))>0)
      cout << " \"" << as_string(car(matches)) << '\"';
    matches = cdr(matches);
  }
  cout << ')';
  cout << DATA_END;
  cout.flush();
}

static void
complete(string text) {
  list matches, tomatch;
  string name;
  tomatch = entries(symbol_table);
  while (!nil(tomatch)) {
    name = as_string(car(tomatch)[0]);
    tomatch = cdr(tomatch);
    if (starts(name,text))
      matches = cons(name(N(text),N(name)), matches);
  }
  output_completion(text,matches);
}

/*
 * texmacs request parser
 */

static int
parse_char(string text, char c, int pos) {
  if (pos >= N(text) || pos < 0) return -1;
  if (text[pos]!=c) return -1;
  return pos+1;
}

static int
parse_blanks(string text, int pos) {
  int i=pos;
  while ((i < N(text)) &&
	 (text[i] == ' ' || text[i] == '\t' || text[i] == '\n'))
    i++;
  return i;
}

static int
parse_uint(string text, int pos, int& p) {
  int i= pos;
  while (i < N(text) && text[i] >= '0' && text[i] <= '9')
    i++;
  p= as_int (text (pos, i));
  return i;
}

static int
parse_identifier(string text, string id, int pos) {
  int i=pos;
  for(int j=0; j<N(id); i++,j++) {
    if (i >= N(text)) return -1;
    if (text[i]!=id[j]) return -1;
  }
  return i;
}

static int
parse_string(string text, int pos, string& s) {
  int esc=0;
  int i=pos;
  if ((i= parse_char(text, '\"', i)) <0) return i;
  while (i < N(text) && (text[i]!='\"' || esc))
    i++;
  s = text(pos+1,i);
  return parse_char(text,'\"',i);
}

static int
parse_completion_request(string text, int pos, string& s, int& p) {
  int i=pos;
  if ((i= parse_char(text, '(', i)) <0) return i;
  if ((i= parse_blanks(text, i)) <0) return i;
  if ((i= parse_identifier(text, "complete", i)) <0) return i;
  if ((i= parse_blanks(text, i)) <0) return i;
  if ((i= parse_string(text, i, s)) <0) return i;
  if ((i= parse_blanks(text, i)) <0) return i;
  if ((i= parse_uint(text, i, p)) <0) return i;
  if ((i= parse_blanks(text, i)) <0) return i;
  if ((i= parse_char(text, ')', i)) <0) return i;
  return i;
}

static void
handle_completion_request(string request) {
  string text;
  int i, p;

  if ((i=parse_completion_request (request, 0, text, p)) < 0 ||
      p<0 || p>N(request)) {
    output_completion ("", list());
    cerr << "mmx warning: ignoring unkown texmacs request";
    return;
  }
  
  i=p-1;
  while (i>=0 && ((text[i]>='a' && text[i]<='z')
		  || (text[i]>='A' && text[i]<='Z')
		  || text[i]=='_'))
    i--;
  complete (text (i+1, p));
}

// readline

void
shell_texmacs_initialize(void) {
  cout << DATA_BEGIN << "verbatim:";
  if (completion_mode) {
    cout << DATA_BEGIN << "command:";
    cout << "(plugin-configure mathemagix (:tab-completion #t))";
    cout << DATA_END;
  }
  cout << DATA_END;
  cout << DATA_BEGIN << "verbatim:";
}

string
shell_texmacs_input (int& line_counter) {
  line_counter++;
  cout << DATA_BEGIN << "channel:prompt" << DATA_END;
  cout << DATA_BEGIN << "latex:" << "{\\red Mmx $\\gg$}" << DATA_END;
  cout << DATA_END;
  fflush (stdout);
  string in;
  string request;
  char c;
  int discard=0;
  while ((c = cin.get())!=EOF) {
    if (c==DATA_COMMAND) {
      discard++;
      continue;
    }
    if (c=='\n' && discard) {
      discard--;
      if (discard==0) {
	if (completion_mode)
	  handle_completion_request(request);
	request = string("");
      }
      continue;
    }
    if (discard) {
      request << c;
      continue;
    }
    if (c=='\n' && !discard) break;
    in << c;
  }
  cout << DATA_BEGIN << "verbatim:";
  if (N(in)==0 && c==EOF) {
    cout << DATA_END;
    cout.flush();
    exit(0);
  }
  return in;
}

} // namespace mmx
