
/******************************************************************************
* MODULE     : gc.cc
* DESCRIPTION: Memory management and garbage collection
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <pointer.hh>
#include <string.hh>
#include <table.hh>
namespace mmx {

static table root_table;
extern table source_location;

static size_t gc_in_use= 0;
static void* last_block= NULL;
static void* needs_unregistration= NULL;

bool interrupt_evaluation= false;
bool gc_debug_flag= false;
static bool gc_busy= false;
static size_t gc_threshold= 16384;

/******************************************************************************
* Registration of entry pointers for the GC
******************************************************************************/

void
gc_register (char* name, void* ptr) {
  root_table (pointer (ptr))= string (name);
  /*
  if (strcmp (name, "glued") == 0)
    cerr << "Registered " << name << " -> "
	 << ptr << ", " << *((generic*) ptr) << lf;
  */
  //if (gc_debug_flag)
  //if (strcmp (name, "glued") == 0)
  //cerr << "Registered " << name << " -> " << ptr << lf;
}

void
gc_unregister (char* name, void* ptr) {
  //cerr << "Unregister " << name << " -> " << *((generic*) ptr) << lf;
  if (gc_busy) {
    void** l= (void**) mmx_malloc (3 * sizeof (void*));
    l[0]= ptr;
    l[1]= (void*) name;
    l[2]= needs_unregistration;
    needs_unregistration= (void*) l;
  }
  else {
    //if (gc_debug_flag)
    /*
    if (strcmp (name, "glued") == 0)
      cerr << "Unregister " << name << " -> " << ptr << lf;
    */
    root_table -> reset (pointer (ptr));
  }
}

static void
delayed_unregistration () {
  while (needs_unregistration != NULL) {
    void** l= (void**) needs_unregistration;
    /*
    char* name= (char*) l[1];
    //if (gc_debug_flag)
    if (strcmp (name, "glued") == 0)
      cerr << "Delayed unregister " << name << " -> " << l[0] << lf;
    */
    root_table -> reset (pointer (l[0]));
    needs_unregistration= l[2];
    mmx_free ((void*) l, 3 * sizeof (void*));
  }
}

void
mmx_register (char* name, void* ptr) {
  //cerr << "Registered " << name << " -> " << *ptr << lf;
}

void
mmx_unregister (char* name, void* ptr) {
  //cerr << "Unregistered " << name << " -> " << *ptr << lf;
}

void
mmx_register_char_array (char* name, char* ptr) {
  //cerr << "Registered " << name << " -> " << ptr << lf;
}

void
mmx_unregister_char_array (char* name, char* ptr) {
  //cerr << "Unregistered " << name << " -> " << ptr << lf;
}

/******************************************************************************
* Marking algorithms
******************************************************************************/

#define gc_info_size sizeof(void*)
#define get_block(object) ((void*) (((char*) (object)) - gc_info_size))
#define get_object(block) ((generic_rep*) (((char*) (block)) + gc_info_size))

#define get_extra(block) ((*((long int*) (block))) & 3)
#define set_ptr(block,ptr) \
  *((long int*) (block))= ((long int) ptr) | get_extra (block);
#define get_ptr(block) ((void*) ((*((long int*) (block))) & (~3)))
#define clear_mark(block) *((long int*) (block)) &= ~1;
#define set_mark(block) *((long int*) (block)) |= 1;
#define is_not_marked(block) (((*((long int*) (block))) & 1) == 0)
#define is_marked(block) ((*((long int*) (block))) & 1)

void
gc_unmark () {
  void* block= last_block;
  while (block != NULL) {
    clear_mark (block);
    block= get_ptr (block);
  }
}

void
gc_mark (generic_rep* object) {
  void* block= get_block (object);
  if (is_not_marked (block)) {
    //cerr << "Mark " << block << " -> " << detailed;
    //cerr << generic (object);
    //cerr << normal << lf << indent;
    set_mark (block);
    object->Mark ();
    //cerr << unindent;
  }
}

void
gc_mark_roots () {
  mark (root_table);
  int i, n= root_table->n;
  for (i=0; i<n; i++) {
    list l= root_table->a[i];
    for (; !nil(l); l= cdr(l)) {
      pointer ptr= as_pointer (car(l)[0]);
      generic_rep* object= ((generic*) ptr->ptr) -> rep;
      //string  name= as_string (car(l)[1]);
      //cerr << lf << hrule;
      //cerr << "Mark " << name << " -> " << ptr->ptr << lf;
      //cerr << "Mark " << name << " -> " << object << lf;
      //cerr << hrule;
      if (object != NULL) gc_mark (object);
    }
  }
  //cerr << hrule;
}

void
gc_mark_sources () {
  mark (source_location);
  int i, n= source_location->n;
  for (i=0; i<n; i++) {
    list l= source_location->a[i];
    for (; !nil(l); l= cdr(l)) {
      pointer ptr= as_pointer (car(l)[0]);
      generic_rep* object= (generic_rep*) ptr->ptr;
      if (object != NULL) gc_mark (object);
    }
  }
}

void
gc_mark () {
  gc_mark_roots ();
  gc_mark_sources ();
}

void
gc_sweep () {
  void* prev_block= NULL;
  void* block= last_block;
  while (block != NULL) {
    void* next_block= get_ptr (block);
    if (is_not_marked (block)) {
      generic_rep* object= get_object (block);
      if (prev_block == NULL) last_block= next_block;
      else set_ptr (prev_block, next_block);
      /*
      cerr << "Destroy " << block << " -> ";
      cerr << generic (object);
      cerr << lf;
      */
      delete object;
      //gc_free (object, object->Size ());
      //cerr << "  Size= " << gc_in_use << "\n";
      //gc_verify_allocated ();
    }
    else prev_block= block;
    block= next_block;
  }
}

void
gc_collect () {
  //size_t old= gc_in_use;
  //cerr << "Launching gc..." << lf;
  gc_busy= true;
  gc_unmark ();
  gc_mark ();
  gc_sweep ();
  delayed_unregistration ();
  gc_busy= false;
  //cerr << "Compressed " << old << " -> " << gc_in_use << lf;
  if (gc_in_use > (1<<26)) gc_threshold= gc_in_use << 1;
  else if (gc_in_use > (1<<21)) gc_threshold= gc_in_use << 2;
  else gc_threshold= gc_in_use << 3;
}

/******************************************************************************
* Allocation
******************************************************************************/

void*
mmx_malloc (size_t sz) {
  //cerr << "    Alloc " << sz << "\n";
  gc_in_use += sz;
  return malloc (sz);
}

void
mmx_free (void* ptr, size_t sz) {
  //cerr << "    Free " << sz << "\n";
  //gc_verify_available (ptr, sz);
  gc_in_use -= sz;
  free (ptr);
}

void*
var_mmx_malloc (size_t sz) {
  //cerr << "Allocate " << sz << "\n";
  //assert (false);
  return mmx_malloc (sz);
}

void
var_mmx_free (void* ptr, size_t sz) {
  //cerr << "Free " << sz << "\n";
  //assert (false);
  mmx_free (ptr, sz);
}

void*
gc_malloc (size_t sz) {
  assert (sz != 0);
  void* block= mmx_malloc (sz + gc_info_size);
  *((void**) block)= last_block;
  last_block= block;
  //gc_verify_allocated ();
  interrupt_evaluation |= (gc_in_use >= gc_threshold);
  return get_object (block);
}

void
gc_free (void* ptr, size_t sz) {
  // Note: it is prohibited to manually delete gc_allocated objects
  mmx_free (get_block (ptr), sz + gc_info_size);
}

/******************************************************************************
* Verification
******************************************************************************/

void
gc_verify_allocated (int n) {
  // verify absence of cycles in list of allocated objects
  void* block= last_block;
  while (block != NULL) {
    block= get_ptr (block);
    n--;
    assert (n>=0);
  }
}

void
gc_verify_available (void* addr, size_t sz) {
  // verify that no already allocated objects starts at address in range
  void* block= last_block;
  while (block != NULL) {
    if (((long) block) >= ((long) addr))
      if (((long) block) < (((long) addr) + ((long) sz))) {
	cerr << "Warning: " << block << " in " << addr << ", " << sz << "\n";
	assert (false);
      }
    block= get_ptr (block);
  }
}

/******************************************************************************
* Inspection
******************************************************************************/

int
mmx_used_bytes () {
  return gc_in_use;
}

int
gc_count_objects (int kind) {
  // 1 -> marked, 2 -> not marked, 3 -> any, negative -> unmark + mark
  if (kind > 0) {
    gc_unmark ();
    gc_mark ();
  }
  else kind= -kind;

  int nr= 0;
  void* block= last_block;
  while (block != NULL) {
    bool count= true;
    if (kind == 1) count= is_marked (block);
    else if (kind == 2) count= is_not_marked (block);
    if (count) nr++;
    block= get_ptr (block);
  }
  return nr;
}

void
gc_show_objects (int kind) {
  // 1 -> marked, 2 -> not marked, 3 -> any, negative -> unmark + mark
  if (kind > 0) {
    gc_unmark ();
    gc_mark ();
  }
  else kind= -kind;

  void* prev_block= NULL;
  void* block= last_block;
  while (block != NULL) {
    bool show= true;
    if (kind == 1) show= is_marked (block);
    else if (kind == 2) show= is_not_marked (block);
    int type= get_object (block)->Type ();
    if (show && ((type == GLUED_TYPE) || (type == GLUED_TYPE_TYPE))) {
      generic_rep* object= get_object (block);
      if (is_marked (block)) cerr << "Marked ";
      else cerr << "Useless ";
      cerr << block << " -> " << detailed << generic (object) << normal << lf;
    }
    prev_block= block;
    block= get_ptr (block);
  }
}

void
gc_inspect () {
  cerr << "Memory usage      : " << gc_in_use << lf;
  cerr << "Objects allocated : " << gc_count_objects (-3) << lf;
  cerr << "Useless objects   : " << gc_count_objects (2) << lf;
}

void
gc_memory_dump () {
  mark (root_table);
  int i, n= root_table->n;
  for (i=0; i<n; i++) {
    list l= root_table->a[i];
    for (; !nil(l); l= cdr(l)) {
      pointer ptr = as_pointer (car(l)[0]);
      string  name= as_string  (car(l)[1]);
      generic_rep* object= ((generic*) ptr->ptr) -> rep;
      cerr << hrule;
      cerr << indented << name << " -> "
	// << expand_cycles << expand_cycles
	   << generic (object)
	// << compress_cycles << compress_cycles
	   << normal << lf;
    }
  }
  cerr << hrule;
}

} // namespace mmx
